#ifndef __LARFLOW_PROJECTION_DEFECT_SPLITTER_H__
#define __LARFLOW_PROJECTION_DEFECT_SPLITTER_H__

#include <vector>
#include <string>
#include <map>

#include "larcv/core/Base/larcv_base.h"
#include "larcv/core/DataFormat/IOManager.h"
#include "larcv/core/DataFormat/Image2D.h"
#include "larlite/DataFormat/storage_manager.h"
#include "larlite/DataFormat/larflowcluster.h"
#include "larlite/DataFormat/larflow3dhit.h"
#include "larlite/DataFormat/track.h"

#include "larflow/RecoUtils/cluster_functions.h"

namespace larflow {
namespace reco {

  /**
   * @ingroup Reco
   * @class ProjectionDefectSplitter
   * @brief Splits 3D clusters by projecting into 2D and breaking cluster based on cluster defect points
   *
   * This is a way to try and produce very straight clusters, which we can then piece 
   * back together to make tracks.
   *
   */
  class ProjectionDefectSplitter : public larcv::larcv_base {
    
  public:

    /** @brief default construct */
    ProjectionDefectSplitter()
      : larcv::larcv_base("ProjectionDefectSplitter"),
      _input_lfhit_tree_name("larmatch"),
      _output_cluster_tree_name("projsplit"),
      _min_larmatch_score(0.0),
      _maxdist(1.0),
      _minsize(20),
      _maxkd(30),
      _kp_veto_radius(3.0),
      _kp_veto_score_threshold(0.85),
      _veto_hits_around_keypoints(false),
      _remove_vetoed_hits(false),
      _cluster_veto_hits(false),
      _fit_line_segments_to_clusters(false)      
      {};
    virtual ~ProjectionDefectSplitter() {};

    void process( larcv::IOManager& iolc, larlite::storage_manager& ioll );

    int split_clusters( std::vector<recoutils::cluster_t>& cluster_v,
                        const std::vector<larcv::Image2D>& adc_v,
                        const float min_second_pca_len );
    
    void doClusterVetoHits( bool doit=true ) { _cluster_veto_hits=doit; };

  protected:

    larlite::larflowcluster
      _makeLArFlowCluster( recoutils::cluster_t& cluster,
                           const larlite::event_larflow3dhit& source_lfhit_v );
    
    recoutils::cluster_t _absorb_nearby_hits( recoutils::cluster_t& cluster,
                                   const std::vector<larlite::larflow3dhit>& hit_v,
                                   std::vector<int>& used_hits_v,
                                   std::vector<larlite::larflow3dhit>& downsample_hit_v,
                                   std::vector<int>& orig_idx_v,                       
                                   float max_dist2line );
    
    void _runSplitter( const larlite::event_larflow3dhit& inputhits,
                       const std::vector<larcv::Image2D>& adc_v,
                       std::vector<int>& used_hits_v,
                       std::vector<recoutils::cluster_t>& output_cluster_v );

    void _findVetoClusters( const larlite::event_larflow3dhit& inputhits,
			    const std::vector<larcv::Image2D>& adc_v,
			    std::vector<int>& used_hits_v,
			    std::vector<recoutils::cluster_t>& output_cluster_v );    
    
    void _defragment_clusters( std::vector<recoutils::cluster_t>& cluster_v,
                               const float max_2nd_pca_eigenvalue );

    void _select_clusters( std::vector<larflow::recoutils::cluster_t>& cluster_v,
			   std::vector<int>& used_hits_v,
			   const std::vector<int>& orig_index_v,
			   const int min_nhits,
			   const float min_second_pca_len );
    
    
    // PARAMETER NAMES
  protected:
    
    std::string _input_lfhit_tree_name;     ///< name of tree to get larflow hits to cluster
    std::string _output_cluster_tree_name;  ///< name of tree to store output clusters
    float _min_larmatch_score;              ///< minimum larmatch score spacepoint must have to be included
    float _maxdist;                         ///< maximum distance two spacepoints can be connected for dbscan
    int   _minsize;                         ///< minimum cluster size for dbscan
    int   _maxkd;                           ///< max number of neighbors per spaecepoint for dbscan
    float _kp_veto_radius;                  ///< radius from keypoint within which spacepoints are vetoed
    float _kp_veto_score_threshold;         ///< keypoint scores above threshold used to veto keypoints

  public:

    /** @brief set name of tree with larflow3dhit instnces to cluster */
    void set_input_larmatchhit_tree_name( std::string name ) { _input_lfhit_tree_name=name; };

    /** @brief set name of the tree to write output clusters */
    void set_output_tree_name( std::string name ) { _output_cluster_tree_name=name; };

    /** @brief set minimum larmatch score must have to be included in clusters */
    void set_min_larmatch_score( float min_score ) { _min_larmatch_score = min_score; };

    /** @brief set keypoint vertex veto radius */
    void set_keypoint_veto_radius( float radius_cm ) { _kp_veto_radius = radius_cm; };

    /** @brief set keypoint score threshold for veto */
    void set_keypoint_score_threshold( float score ) { _kp_veto_score_threshold = score; };
    

    /** 
     * @brief set the dbscan parameters for clustering 
     * @param[in] maxdist maximum distance two spacepoints can be connected for dbscan
     * @param[in] minsize minimum cluster size for dbscan
     * @param[in] maxkd   max number of neighbors per spaecepoint for dbscan
     */
    void set_dbscan_pars( float maxdist, int minsize, int maxkd ) {
      _maxdist = maxdist;
      _minsize = minsize;
      _maxkd = maxkd;
    };

    // OPTIONAL INPUTS

  public:
    
    void add_input_keypoint_treename_for_hitveto( std::string name );
    
  protected:

    bool _veto_hits_around_keypoints; ///< if true, veto hits around keypoints to help separate particle clusters
    bool _remove_vetoed_hits;         ///< if true, we remove veto hits, else we push away from the keypoint
    bool _cluster_veto_hits;          ///< if true, we try to cluster veto'd hits around the keypoint
    std::vector< std::string > _keypoint_veto_trees_v; ///< contains name of keypoint tree names for vetoing hits
    std::vector< const larlite::event_larflow3dhit* > _event_keypoint_for_veto_v;  ///< we veto clusters if they are near these 3d space points
    std::vector< const larlite::larflow3dhit* > _keypoints_for_veto_v;
    std::map<int,int>  _hitidx_to_kpidx;  ///< for vetoed hits, the keypoint index it was vetoed by
    std::vector<float> _hitidx_to_kpdist; ///< distance to keypoint
    int _veto_hits_using_keypoints( const larlite::event_larflow3dhit& inputhits,
                                    std::vector<int>& used_hits_v );
    int _veto_hits_using_keypoint_scores( const larlite::event_larflow3dhit& inputhits,
					  std::vector<int>& used_hits_v,
					  float kp_score_threshold );

    bool _fit_line_segments_to_clusters; ///< if true, run routine to fit line segments to the different clusters

  public:
    
    static void fitLineSegmentsToClusters( const std::vector<larflow::recoutils::cluster_t>& cluster_v,
                                           const larlite::event_larflow3dhit& lfhit_v,
                                           const std::vector<larcv::Image2D>& adc_v,
                                           larlite::event_track& evout_track );
    
    static larlite::track fitLineSegmentToCluster( const larflow::recoutils::cluster_t& cluster,
                                                   const larlite::event_larflow3dhit& lfhit_v,
                                                   const std::vector<larcv::Image2D>& adc_v,
                                                   const float max_line_seg_cm=5.0);
  public:

    /** @brief set the flag that determines if linesegments are fit to the clusters */
    void set_fit_line_segments_to_clusters( bool fit ) { _fit_line_segments_to_clusters=fit; };
    
    
    
  };

}
}

#endif
